import logging
import logging
import time
from asyncio import get_event_loop
from typing import Optional

from .. import util, create_irc
from ..channel import Channel, channels
from ..command import Command, commands, CustomCommandAction, is_command_on_cooldown, get_time_since_execute, \
    update_command_last_execute
from ..config import cfg, get_nick
from ..config import generate_config
from ..database import get_custom_command
from ..disabled_commands import is_command_disabled
from ..emote import update_global_emotes
from ..enums import Event
from ..enums import MessageType, CommandContext
from ..events import trigger_event
from ..exceptions import InvalidArgumentsError
from ..irc import Irc
from ..message import Message
from ..modloader import Mod
from ..modloader import mods
from ..modloader import trigger_mod_event
from ..permission import perms
from ..shared import set_bot
from ..util import stop_all_tasks
from ..command_whitelist import is_command_whitelisted, send_message_on_command_whitelist_deny


# noinspection PyMethodMayBeStatic
class BaseBot:
    def __init__(self):
        self.irc: Irc = None
        self._running = False
        set_bot(self)

    # region events
    async def on_bot_banned_from_channel(self, msg: Message, channel: Channel):
        """
        triggered when the bot attempts to join a banned channel
        :param msg: the message that twitch sent saying the bot was banned
        :param channel: the channel the bot was banned from
        """

    async def on_bot_timed_out_from_channel(self, msg: Message, channel: Channel, seconds: int):
        """
        triggered when the bot is timed out on a channel
        :param msg: the message that twitch sent saying the bot was timed out
        :param channel: the channel the bot was timed out on
        :param seconds: how many seconds left in the timeout
        """

    async def on_mod_reloaded(self, mod: Mod):
        """
        triggered when a mod is reloaded using reload_mod() or !reloadmod
        :param mod: mod being reloaded
        """

    async def on_connected(self):
        """
        triggered when the bot connects to all the channels specified in the config file
        """

    async def on_raw_message(self, msg: Message):
        """
        triggered the instant a message is received,
        this message can be any message received,
        including twitches messages that do not have any useful information
        """

    async def on_privmsg_sent(self, msg: str, channel: str, sender: str) -> None:
        """
        triggered when the bot sends a privmsg
        """
        print(f'{sender}({channel}): {msg}')

    async def on_privmsg_received(self, msg: Message) -> None:
        """triggered when a privmsg is received, is not triggered if the msg is a command"""

    async def on_whisper_sent(self, msg: str, receiver: str, sender: str):
        """
        triggered when the bot sends a whisper to someone
        """
        print(f'{sender} -> {receiver}: {msg}')

    async def on_whisper_received(self, msg: Message):
        """
        triggered when a user sends the bot a whisper
        """

    async def on_permission_check(self, msg: Message, cmd: Command) -> bool:
        """
        triggered when a command permission check is requested
        :param msg: the message the command was found from
        :param cmd: the command that was found
        :return: bool indicating if the user has permission to call the command, True = yes, False = no
        """
        return True

    async def on_before_command_execute(self, msg: Message, cmd: Command) -> bool:
        """
        triggered before a command is executed
        :return bool, if return value is False, then the command will not be executed
        """
        return True

    async def on_after_command_execute(self, msg: Message, cmd: Command) -> None:
        """
        triggered after a command has executed
        """

    async def on_bits_donated(self, msg: Message, bits: int):
        """
        triggered when a bit donation is posted in chat
        """

    async def on_channel_raided(self, channel: Channel, raider: str, viewer_count: int):
        """
        triggered when the channel is raided
        :param channel: the channel who was raided
        :param raider: the user who raided
        :param viewer_count: the number of viewers who joined in the raid
        """

    async def on_channel_joined(self, channel: Channel):
        """
        triggered when the bot joins a channel
        """
        print(f'joined #{channel.name}')

    async def on_channel_points_redemption(self, msg: Message, reward: str):
        """
        triggered when a viewers redeems channel points for a reward
        """
        print(f'{msg.author} has redeemed channel points reward "{reward}" in #{msg.channel_name}')

    async def on_user_join(self, user: str, channel: Channel):
        """
        triggered when a user joins a channel the bot is in
        :param user: the user who joined
        :param channel: the channel that the user joined
        """

    async def on_user_part(self, user: str, channel: Channel):
        """
        triggered when a user leaves from a channel the bot is in
        :param user: the user who left
        :param channel: the channel that the user left
        """

    async def on_channel_subscription(self, subscriber: str, channel: Channel, msg: Message):
        """
        triggered when a user subscribes
        """

    # endregion

    def _create_channels(self):
        for name in cfg.channels:
            chan = Channel(name, irc=self.irc)
            chan.start_update_loop()

    async def _create_irc(self):
        """
        creates the async reader/writer (using asyncio.open_connection() if not already exist),
        """
        self.irc = await create_irc()

    def _request_permissions(self):
        """requests permissions from twitch to be able to gets message tags, receive whispers, ect"""
        # enable receiving/sending whispers
        self.irc.send('CAP REQ :twitch.tv/commands')

        # enable seeing bit donations and such
        self.irc.send('CAP REQ :twitch.tv/tags')

        # enable seeing user joins
        self.irc.send('CAP REQ :twitch.tv/membership')

    async def _connect(self):
        """connects to twitch, sends auth info, and joins the channels in the config"""
        print(f'logging in as {get_nick()}')

        util.send_auth(self.irc)

        resp = (await self.irc.get_next_message()).lower()
        if 'authentication failed' in resp:
            print(
                '\n\n=========AUTHENTICATION FAILED=========\n\n'
                'check that your oauth is correct and valid and that the nick in the config is correct'
                '\nthere is a chance that oauth was good, but is not anymore\n'
                'the oauth token can be regenerated using this website: \n\n\thttps://twitchapps.com/tmi/')
            input('\n\npress enter to exit')
            exit(1)
        elif 'welcome' not in resp:
            print(
                f'\n\ntwitch gave a bad response to sending authentication to twitch server\nbelow is the message received from twitch:\n\n\t{resp}')
            input('\n\npress enter to exit')
            exit(1)

        self._request_permissions()

        for chan in channels.values():
            self.irc.send(f'JOIN #{chan.name}')

    async def get_command_from_msg(self, msg: Message) -> Optional[Command]:
        """
        checks if the start of the msg matches any command names

        if command is found: returns the command

        else: returns None
        """
        cmd = commands.get(msg.parts[0].lower())
        if cmd:
            return cmd

        cmd = get_custom_command(msg.channel_name, msg.parts[0].lower())
        if cmd:
            return CustomCommandAction(cmd)

        return None

    async def _run_command(self, msg: Message, cmd: Command):
        if (
                not await self.on_permission_check(msg, cmd)
                or not all(await trigger_mod_event(Event.on_permission_check, msg, cmd, channel=msg.channel_name))
                or not all(await trigger_event(Event.on_permission_check, msg, cmd))
        ):
            return await msg.reply(
                whisper=True,
                msg=f'you do not have permission to execute {cmd.fullname} in #{msg.channel_name}, permission required: {cmd.permission}')

        elif not isinstance(cmd, CustomCommandAction) and is_command_disabled(msg.channel_name, cmd.fullname):
            return await msg.reply(f'{cmd.fullname} is disabled for this channel')

        # also check if the command is whitelisted, (if its not a custom command)
        if not isinstance(cmd, CustomCommandAction) and not is_command_whitelisted(cmd.name):
            if send_message_on_command_whitelist_deny():
                await msg.reply(f'{msg.mention} "{cmd.fullname}" is not enabled in the command whitelist')
            return

        has_cooldown_bypass_permission = perms.has_permission(msg.channel_name, msg.author, cmd.cooldown_bypass)
        if (not has_cooldown_bypass_permission
                and is_command_on_cooldown(msg.channel_name, cmd.fullname, cmd.cooldown)):
            return await msg.reply(
                f'{cmd.fullname} is on cooldown, seconds left: {cmd.cooldown - get_time_since_execute(msg.channel_name, cmd.fullname)}')

        if (not await self.on_before_command_execute(msg, cmd)
                or not all(await trigger_mod_event(Event.on_before_command_execute, msg, cmd, channel=msg.channel_name))
                or not all(await trigger_event(Event.on_before_command_execute, msg, cmd))
        ):
            return

        try:
            await cmd.execute(msg)
            if not has_cooldown_bypass_permission:
                update_command_last_execute(msg.channel_name, cmd.fullname)
        except InvalidArgumentsError as e:
            await self._send_cmd_help(msg, cmd, e)
        else:
            await self.on_after_command_execute(msg, cmd)
            await trigger_mod_event(Event.on_after_command_execute, msg, cmd, channel=msg.channel_name)
            await trigger_event(Event.on_after_command_execute, msg, cmd)

    async def _send_cmd_help(self, msg: Message, cmd: Command, exc: InvalidArgumentsError):
        await msg.reply(
            f'{exc.reason} - "{cmd.fullname} {cmd.syntax}" - do "{cfg.prefix}help {cmd.fullname}" for more details')

    # kept if needed later
    # def _load_overrides(self):
    #     for k, v in overrides.items():
    #         if k.value in self.__class__.__dict__ and k.value.startswith('on'):
    #             setattr(self, k.value, v)

    def shutdown(self):
        for channel in channels:
            self.irc.send(f'PART #{channel}')
            time.sleep(.4)
        self.irc.send('QUIT')
        self._running = False
        stop_all_tasks()

    def run(self):
        """runs/starts the bot, this is a blocking function that starts the mainloop"""
        self._running = True
        get_event_loop().run_until_complete(self.mainloop())

    async def mainloop(self):
        """starts the bot, connects to twitch, then starts the message event loop"""
        # check if user wants to input oauth info manually
        if not generate_config():
            stop_all_tasks()
            return

        await update_global_emotes()

        await self._create_irc()
        self._create_channels()

        await self._connect()
        await self.on_connected()
        await trigger_mod_event(Event.on_connected)
        await trigger_event(Event.on_connected)

        while self._running:
            raw_msg = await self.irc.get_next_message()

            if not raw_msg:
                continue

            msg = Message(raw_msg, irc=self.irc, bot=self)

            await self.on_raw_message(msg)
            get_event_loop().create_task(trigger_mod_event(Event.on_raw_message, msg, channel=msg.channel_name))
            get_event_loop().create_task(trigger_event(Event.on_raw_message, msg))

            coro = mod_coro = event_coro = None
            cmd: Command = (await self.get_command_from_msg(msg)
                            if msg.is_user_message and msg.author != get_nick()
                            else None)

            if cmd and ((msg.is_whisper and cmd.context & CommandContext.WHISPER)
                        or (msg.is_privmsg and cmd.context & CommandContext.CHANNEL)):
                msg.safe_print()
                coro = self._run_command(msg, cmd)

            elif msg.type is MessageType.WHISPER:
                msg.safe_print()
                coro = self.on_whisper_received(msg)
                mod_coro = trigger_mod_event(Event.on_whisper_received, msg)
                event_coro = trigger_event(Event.on_whisper_received, msg)

            elif msg.type is MessageType.PRIVMSG:
                msg.safe_print()
                coro = self.on_privmsg_received(msg)
                mod_coro = trigger_mod_event(Event.on_privmsg_received, msg, channel=msg.channel_name)
                event_coro = trigger_event(Event.on_privmsg_received, msg)

            elif msg.type is MessageType.USER_JOIN:
                # the bot has joined a channel
                if msg.author == get_nick():
                    coro = self.on_channel_joined(msg.channel)
                    mod_coro = trigger_mod_event(Event.on_channel_joined, msg.channel, channel=msg.channel_name)
                    event_coro = trigger_event(Event.on_channel_joined, msg.channel)
                # user joined a channel the bot was in
                else:
                    coro = self.on_user_join(msg.author, msg.channel)
                    mod_coro = trigger_mod_event(Event.on_user_join, msg.author, msg.channel,
                                                 channel=msg.channel_name)
                    event_coro = trigger_event(Event.on_user_join, msg.author, msg.channel)

            elif msg.type is MessageType.USER_PART:
                coro = self.on_user_part(msg.author, msg.channel)
                mod_coro = trigger_mod_event(Event.on_user_part, msg.author, msg.channel, channel=msg.channel_name)
                event_coro = trigger_event(Event.on_user_part, msg.author, msg.channel)

            elif msg.type is MessageType.SUBSCRIPTION:
                coro = self.on_channel_subscription(msg.author, msg.channel, msg)
                mod_coro = trigger_mod_event(Event.on_channel_subscription, msg.author, msg.channel, msg,
                                             channel=msg.channel_name)
                event_coro = trigger_event(Event.on_channel_subscription, msg.author, msg.channel, msg)

            elif msg.type is MessageType.RAID:
                coro = self.on_channel_raided(msg.channel, msg.author, msg.tags.raid_viewer_count)
                mod_coro = trigger_mod_event(Event.on_channel_raided, msg.channel, msg.author,
                                             msg.tags.raid_viewer_count,
                                             channel=msg.channel_name)
                event_coro = trigger_event(Event.on_channel_raided, msg.channel, msg.author,
                                           msg.tags.raid_viewer_count)

            elif msg.type is MessageType.PING:
                self.irc.send_pong()

            elif msg.type is MessageType.CHANNEL_POINTS_REDEMPTION:
                coro = self.on_channel_points_redemption(msg, msg.reward)
                mod_coro = trigger_mod_event(Event.on_channel_points_redemption, msg, msg.reward,
                                             channel=msg.channel_name)
                event_coro = trigger_event(Event.on_channel_points_redemption, msg, msg.reward)

            elif msg.type is MessageType.BITS:
                coro = self.on_bits_donated(msg, msg.tags.bits)
                mod_coro = trigger_mod_event(Event.on_bits_donated, msg, msg.tags.bits, channel=msg.channel_name)
                event_coro = trigger_event(Event.on_bits_donated, msg.channel, msg)

            elif msg.type is MessageType.BOT_PERMANENTLY_BANNED:
                coro = self.on_bot_banned_from_channel(msg, msg.channel)
                mod_coro = trigger_mod_event(Event.on_bot_banned_from_channel, msg, msg.channel, channel=msg.channel_name)
                event_coro = trigger_event(Event.on_bot_banned_from_channel, msg, msg.channel)

            elif msg.type is MessageType.BOT_TIMED_OUT:
                coro = self.on_bot_timed_out_from_channel(msg, msg.channel, msg.timeout_seconds)
                mod_coro = trigger_mod_event(Event.on_bot_timed_out_from_channel, msg, msg.channel, msg.timeout_seconds, channel=msg.channel_name)
                event_coro = trigger_event(Event.on_bot_timed_out_from_channel, msg, msg.channel, msg.timeout_seconds)

            if coro is not None:
                get_event_loop().create_task(coro)

            if mod_coro is not None:
                get_event_loop().create_task(mod_coro)

            if event_coro is not None:
                get_event_loop().create_task(event_coro)

        # clean up mods when the bot is exiting
        for mod in mods.values():
            # notify all mods of being unloaded,
            # this is put in a try/except
            # so that any exceptions raised from unloaded overrides will not cancel unloading the others
            try:
                await mod.unloaded()
            except Exception as e:
                print(f'\nwhen unloading mod "{mod.name}" this exception occurred:\n')
                logging.exception(e)



from asyncio import sleep, get_event_loop

import typing
from collections import defaultdict
from datetime import datetime

if typing.TYPE_CHECKING:
    from .channel import Channel

__all__ = [
    'PRIVMSG_MAX_MOD',
    'PRIVMSG_MAX_NORMAL',
    'WHISPER_MAX',

    'privmsg_ratelimit',
    'privmsg_sent',
    'privmsg_sent_reset_loop',

    'whisper_ratelimit',
    'whisper_sent',
    'whisper_sent_reset_loop',
]

PRIVMSG_MAX_MOD = 100
PRIVMSG_MAX_NORMAL = 20
PRIVMSG_INTERVAL = 1

WHISPER_MAX = 10

privmsg_sent = 0
whisper_sent = 0

last_privmsg_sent_dt: typing.DefaultDict[str, datetime] = defaultdict(lambda: datetime.min)


def _get_last_privmsg_send_dt(channel: 'Channel') -> float:
    return (datetime.now() - last_privmsg_sent_dt[channel.name]).total_seconds()


async def privmsg_ratelimit(channel: 'Channel'):
    from .config import get_nick
    global privmsg_sent
    limit = PRIVMSG_MAX_NORMAL

    use_mod_limit = channel.is_mod or channel.is_vip or channel.name == get_nick()
    if use_mod_limit:
        limit = PRIVMSG_MAX_MOD

    while not use_mod_limit and _get_last_privmsg_send_dt(channel) < PRIVMSG_INTERVAL:
        await sleep(.3)

    while privmsg_sent >= limit:
        await sleep(1)

    privmsg_sent += 1
    last_privmsg_sent_dt[channel.name] = datetime.now()


async def whisper_ratelimit():
    global whisper_sent

    while whisper_sent >= WHISPER_MAX:
        await sleep(1)

    whisper_sent += 1


async def privmsg_sent_reset_loop():
    global privmsg_sent

    while True:
        privmsg_sent = 0
        await sleep(30)


async def whisper_sent_reset_loop():
    global whisper_sent

    while True:
        whisper_sent = 0
        await sleep(1)


get_event_loop().create_task(privmsg_sent_reset_loop())
get_event_loop().create_task(whisper_sent_reset_loop())



import asyncio
from asyncio import Future, TimeoutError
from typing import List, Callable, Tuple, Awaitable

from .message import Message

__all__ = [
    'reply_wait_queue',
    'ReplyWaitType',
    'same_author_predicate',
    'same_channel_predicate',
    'wait_for_reply',
    'custom_predicate',
    'custom_async_predicate',
]

ReplyWaitType = Tuple[Future, Callable[..., Awaitable[bool]]]

# list of futures and coroutines that are waiting for a message reply
# first item in each tuple must be a Future, this is because it is later given a value if the Awaitable return True
# second item in each tuple must be a Awaitable that return True or False
# a True return from the Awaitable means the message was accepted, and the Future has its result set to the message
reply_wait_queue: List[ReplyWaitType] = []


def same_author_predicate(msg: Message):
    """
    returns a async predicate where the message from be from the same author and channel as `msg` passed to this function
    """

    async def _same_author_predicate(m):
        return m.channel == msg.channel and m.author == msg.author and m.content != msg.content

    return _same_author_predicate


def same_channel_predicate(msg: Message):
    """
    returns a async predicate where the message from be from the same channel as `msg` passed to this function
    """

    async def _same_channel_predicate(m):
        return m.channel == msg.channel

    return _same_channel_predicate


def custom_predicate(custom_predicate: Callable[[Message], bool] = None,
                     same_author=True,
                     same_channel=True,
                     msg: Message = None):
    """
    returns a async predicate where the custom_predicate must return True when called with the message

    if same_author is True, then the message must come from the same author who sent msg

    if same_channel is True, then the message must come from the same channel as msg was sent from
    """

    if not msg and any((same_author, same_channel)):
        raise ValueError(
            'msg cannot be None if same_author or same_channel is True, add `msg=MSG_HERE` to fix this error')

    async def _custom_predicate(m: Message):
        if (same_channel and m.channel != m.channel) or (same_author and m.author != msg.author):
            return False

        return custom_predicate(m)

    return _custom_predicate


def custom_async_predicate(msg: Message, custom_predicate: Callable[[Message], Awaitable[bool]] = None,
                           same_author=True,
                           same_channel=True):
    async def _custom_async_predicate(m: Message):
        if (same_channel and m.channel != m.channel) or (same_author and m.author != msg.author):
            return False
        return await custom_predicate(m)

    return _custom_async_predicate


async def wait_for_reply(predicate: Callable[[Message], Awaitable[bool]] = None, timeout=30, default=None,
                         raise_on_timeout=False) -> Message:
    """
    waits for a message matching `predicate` to be received, when its received, it returns that message.

    if no message matching `predicate` is received by the timeout, the default will be returned.

    if raise_on_timeout is True and no matching message is received, this function will raise asyncio.TimeoutError

    if raise_on_timeout is False and no matching message is received, default will be returned when it times-out

    default is by default is None
    raise_on_timeout by default is False
    """

    async def _timeout_defaulter():
        try:
            return await asyncio.wait_for(future, timeout)
        except TimeoutError:
            if raise_on_timeout:
                raise
            else:
                return default

    future = asyncio.get_event_loop().create_future()
    reply_wait_queue.append((future, predicate))
    return await _timeout_defaulter()



import asyncio

from twitchbot import Command, CommandContext, Message, get_bot, stop_all_tasks

ADMIN_COMMAND_PERMISSION = 'admin'


@Command('shutdown', aliases=['stop', 's'], context=CommandContext.BOTH, permission=ADMIN_COMMAND_PERMISSION,
         help='make the bot shutdown')
async def cmd_shutdown(msg: Message, *args):
    await msg.reply('bot shutting down...')
    stop_all_tasks()

    for seconds_left in range(10, 0, -1):
        print(f'giving running tasks time to stop, closing in {seconds_left} seconds...')
        await asyncio.sleep(1)

    get_bot().shutdown()



import shlex
import typing

from typing import Union
from ..regex import RE_AT_MENTION

__all__ = ('split_message', 'get_message_mentions')

# little "hack" to get around circular imports for type hinting
if typing.TYPE_CHECKING:
    from ..message import Message


def split_message(msg: str):
    try:
        return shlex.split(msg)
    except ValueError:
        return msg.split(' ')


def get_message_mentions(message: Union['Message', str]):
    # hack to get around circular imports
    from ..message import Message

    # try getting the attribute "content", if it fails, text is set to the passed message itself
    # which should be a string
    text = getattr(message, 'content', message)
    mentions = tuple(RE_AT_MENTION.findall(text))

    # checks for username mentions without the @
    if isinstance(message, Message):
        mentions += tuple(p for p in message.parts if p in message.channel.chatters)

    return mentions



from .util import add_task, task_running, stop_task
from asyncio import sleep
from .database import add_balance, session
from .config import cfg
from .channel import channels
from .config import cfg

LOYALTY_TICKER_TASK_NAME = '_loyalty_ticker'


async def _ticker_loop():
    while True:
        for channel in channels.values():
            for viewer in channel.chatters.all_viewers:
                add_balance(channel.name, viewer, cfg.loyalty_amount, commit=False)
                await sleep(.2)
            session.commit()

        await sleep(cfg.loyalty_interval)


def start_loyalty_ticker():
    if cfg.loyalty_interval == -1:
        print('loyalty_interval in config is -1, loyalty ticker WILL NOT start')
        return

    if not task_running(LOYALTY_TICKER_TASK_NAME):
        add_task(LOYALTY_TICKER_TASK_NAME, _ticker_loop())


def stop_loyalty_ticker():
    stop_task(LOYALTY_TICKER_TASK_NAME)


start_loyalty_ticker()
